package com.ibm.ctg.ping;

/**
 *   Class: Pinger
 *   
 *   Support pack for IBM CICS Transaction Gateway Version v5.0 to v7.2 
 *
 *   Version       : 1.3
 *
 *   Description   : Connects to a specified Gateway daemon, issues requests
 *   and measures the time taken for operations to complete. This is useful for 
 *   testing and debugging connectivity and performance.
 *
 *   See support pack documentation for details on how to configure and use this
 *   sample. Also see the support pack for information on configuration necessary on
 *   CICS.
 *   
 *   Licensed Materials - Property of IBM
 *
 *   Copyright IBM Corporation 2008
 *   All Rights Reserved
 *
 *   US Government Users Restricted Rights - Use,
 *   duplication, or disclosure restricted by GSA
 *   Schedule Contract with IBM Corporation
 *
 *
 *   The following code is sample code created by IBM Corporation.  This
 *   sample code is not part of any standard IBM product and is provided
 *   to you solely for the purpose of assisting you in the development of
 *   your applications.  The code is provided 'AS IS', without warranty
 *   or condition of any kind.  IBM shall not be liable for any damages
 *   arising out of your use of the sample code, even if IBM has been
 *   advised of the possibility of such damages.
 * 
 */

import java.io.IOException;

import com.ibm.ctg.client.ECIRequest;
import com.ibm.ctg.client.JavaGateway;
import com.ibm.ctg.client.exceptions.ContainerExistsException;
import com.ibm.ctg.client.exceptions.ContainerNotFoundException;

public class Pinger extends CTGping implements Constants {

   /**
    * Used to display output
    */
   private DisplayOut displayOutput = null;

   /**
    * The 'think time' between ping requests within an iteration.
    */
   private int interval = DEFAULT_INTERVAL;

   /**
    * The 'think time' between iterations.
    */
   // private int thinkTime = DEFAULT_INTERVAL;
   public ThinkTimeArgument thinkTime = new ThinkTimeArgument('d',
         DEFAULT_INTERVAL, Argument.VALUE_REQUIRED);

   /*
    * The following variables are set, or overwridden, by ParseArgs(). They
    * control behaviour specified by the user as command line parameters.
    */

   /**
    * The number of iterations of open/ping/close requests.
    */
   // private int iterations = DEFAULT_ITERATIONS;
   public Argument iterations = new Argument('i', DEFAULT_ITERATIONS,
         Argument.VALUE_REQUIRED);

   /**
    * The number of requestes within each iteration.
    */
   // public int numberRequests = DEFAULT_REQUESTS;
   public Argument numberRequests = new Argument('r', DEFAULT_REQUESTS,
         Argument.VALUE_REQUIRED);

   /**
    * The URL of the Gateway daemon to connect to.
    */
   public URLSpecifier gatewayURL = null;

   /**
    * Set to the name of the CICS region if the user specifed they wanted to
    * ping CICS.
    */
   // private String cicsRegion = null;
   public Argument cicsRegion = new Argument('c', null,
         Argument.VALUE_NOT_REQUIRED);

   // private String userid = null;
   Argument userid = new Argument('u', null, Argument.STRING_ARG,
         Argument.VALUE_REQUIRED);

   // private String password = null;
   Argument password = new Argument('p', null, Argument.STRING_ARG,
         Argument.VALUE_REQUIRED);

   private Argument help = new Argument('?', "", Argument.VALUE_NOT_REQUIRED);

   // private boolean passwordSpecified = false;

   // private int caLen = DEFAULT_CALEN;
   PayloadArgument caLen = new PayloadArgument('l', DEFAULT_CALEN,
         Argument.VALUE_REQUIRED);

   // private boolean lengthSpecified;

   private int multiplier = NONE;

   private Argument arguments[];


   FlexiRequest pingRequest;

   private Pinger() {
   }

   public Pinger(DisplayOut displayOutput) {
      super();
      this.displayOutput = displayOutput;
   }

   public int runMain(String[] args) {
      ECIRequest runTxn = null;
      byte[] myPayload = null;
      int rc = RC_OK;

      if (args.length == 0) {
         rc = INVALID_PARAMETERS;
      } else {
         rc = parseArgs(args);
      }

      // In order to give equal weighting to each request in the sequence
      // and reduce the impact that this application has on the measurement
      // of the Gateway daemon and CICS performance we do some class loading
      // and object creation upfront, before any requests are issued.
      if (rc == RC_OK) {
         warmUp();
         if (cicsRegion.isSpecified()) {
            // Create a base transaction request. All request types extend ECIRequest. 
            // The actual object created depends on the type of request.
            // This was implemented so that the ctgping utility can be run against 
            // versions of the CICS TG which support channels and containers and versions
            // which don't without producing a class load error. The majority of applications
            // are not required to run against multiple versions of the CICS TG and would 
            // directly create an ECIRequest object.
            pingRequest = FlexiRequestCreator.createCICSRequest(userid
                  .getValueAsString(), password.getValueAsString(), cicsRegion
                  .getValueAsString(), caLen.getValueAsInt(), caLen
                  .isChannelsAndContainers());
            if (pingRequest == null) {
               rc = UNEXPECTED_ERROR;
            } else {
               myPayload = createPayload(caLen.getValueAsInt());
               rc = pingRequest.addPayloadToRequest(myPayload);
               runTxn = pingRequest;
            }
         } else {
            runTxn = ECIRequest.listSystems(LIST_SYSTEMS_MAX_SYSTEMS);
            runTxn.Userid = userid.getValueAsString();
            runTxn.Password = password.getValueAsString();
         }

      }

      if (rc == RC_OK) {
         try {
            rc = runPings(runTxn, myPayload);
         } catch (Exception e) {
            e.printStackTrace();
            rc = UNEXPECTED_ERROR;
         }
      }

      if (rc == UNEXPECTED_ERROR) {
         displayOutput.unexpectedError();
      }

      return rc;
   }

   /**
    * Ensure that the first request flowed is not artificially large by ensuring
    * that it does not invoke class loading for the first time.
    * 
    */
   private static void warmUp() {
      // Open a dummy gateway object so class loader delays are not reflected
      // in the ping time.
      JavaGateway dummy = new JavaGateway();
      try {
         dummy.setURL("tcp://0.0.0.0:9999");
         dummy.open();
      } catch (IOException e1) {
         // Ignore IOException as we expect it
         // System.out.println(e1);
      }
      return;
   }

   public int parseArgs(String argv[]) {

      int rc = RC_OK;
      arguments = new Argument[] { thinkTime, iterations, numberRequests,
            cicsRegion, userid, password, caLen, help };

      Parser myCommand = new Parser(arguments);
      boolean result = myCommand.parseArgs(argv);
      if (help.isSpecified()) {
         rc = HELP_SPECIFIED;
      }
      if (rc == RC_OK) {
         if (result) {
            if (testForInvalidCombinations()) {
               rc = INVALID_PARAMETERS;
            }
         } else {
            rc = INVALID_PARAMETERS;
         }
         if (rc == RC_OK) {
            rc = parseURL(myCommand.getTargetURL());
         }
      } else {
         // Failed to parse the command line successfully
         rc = INVALID_PARAMETERS;
      }
      return rc;
   }

   private boolean testForInvalidCombinations() {
      boolean rc = false;
      // We don't allow people to specify a password without a userid
      if (password.isSpecified() && !userid.isSpecified()) {
         System.out
               .println("A userid must be specified when a password is specified");
         rc = true;
      }

      if (caLen.isSpecified() && !cicsRegion.isSpecified()) {
         System.out
               .println("-l requires that the -c parameter is also specified");
         rc = true;
      }
      return rc;
   }

   /**
    * The main ping logic. Opens connections and issues runTxn requests against
    * them. All requests are timed and the results written to the console.
    * 
    * For demonstration purposes, this method is deliberately left large so that it includes
    * the CICS TG API calls in-line. This is to demontrate the order that the API calls need to be 
    * invoked, and the function of each call.  
    * 
    * @param runTxn
    *           the ping request to use. This will either be set to a
    *           ListSystems call or an ECI request.
    * @throws Exception
    */
   public int runPings(ECIRequest runTxn, byte myPayload[]) throws IOException {
      int rc = RC_OK;
      RunningTotal openTimings = new RunningTotal(MIN_TIME);
      RunningTotal closeTimings = new RunningTotal(MIN_TIME);
      RunningTotal requestTimings = new RunningTotal(MIN_TIME);
      boolean opened = false;

      displayOutput.displayTitleInformation(this);
      //Create a JavaGateway instance. This is used to connect an application
      //to a Gateway daemon.
      JavaGateway jg = null;
      try {
         jg = new JavaGateway();        
         jg.setURL(gatewayURL.getURL());
         if (gatewayURL.isPortSpecified()) {
            try {
               jg.setPort(gatewayURL.getPort());
            } catch (PortNotSpecifiedException e) {
               // We should never get to this point.
               // Store the error string and fail the method if we do.
               ErrorStore.getInstance().set(e);
               throw new IOException(e.getMessage());
            }
         }
      } catch (IOException e) {
         // An invalid address was specified on the Gateway. Treat this as an
         // open fail type error.
         e.printStackTrace();
         ErrorStore.getInstance().set(e);
         rc = OPEN_FAIL;
      }

      if (rc == RC_OK) {
         for (int i = iterations.getValueAsInt(); i > 0; i--) {
            openTimings.setStart();
            try {
               //We open and close a connection for every iteration, in order to
               //measure how long an open takes. Pooling connections and keeping 
               //them open is recommended when performance is a concern.
               jg.open();
               rc = RC_OK;
            } catch (Exception e) {
               openTimings.fail();
               ErrorStore.getInstance().set(e);
               rc = OPEN_FAIL;
            }
            openTimings.setEnd();
            
            if (rc == RC_OK) {
               openTimings.success();
            }
            
            if (rc == RC_OK) {
               displayOutput.successfulOpen(openTimings, gatewayURL);
               opened = true;
            } else {
               displayOutput.failedOpen(gatewayURL);
            }

            if (rc == RC_OK) {
               for (int requestLoop = 0; requestLoop < numberRequests
                     .getValueAsInt(); requestLoop++) {
                  if (caLen.isChannelsAndContainers()) {
                     // Make sure a  channel and container is re-transmitted
                     // for each test. This step is only necessary for v7.1 
                     // of the CICS TG.
                     pingRequest.addPayloadToRequest(myPayload);
                  }
                  
                  requestTimings.setStart();
                  try {
                     //Send a request over the connection.
                     //If runTxn is a list systems ECIRequest call, this will cause the
                     //Gateway daemon to return the number of configured CICS servers.
                     //If runTxn is an ECIRequest with a program specified, the Gateway
                     //daemon will flow the request to CICS, and send the response back to
                     //us.
                     jg.flow(runTxn);
                  } catch (IOException e) {
                     //An IOException generally indicates that a networking error occurred
                     //between the application and the Gateway. 
                     rc = RC_REQUEST_FAIL;
                     ErrorStore.getInstance().set(e);
                  }
                  requestTimings.setEnd();
                                
                  try {
                     Thread.sleep(interval);
                  } catch (InterruptedException e) {
                     e.printStackTrace();
                  }

                  //If an IOException did not occur, we need to look at the return codes
                  //from the request object itself to verify that it was correctly 
                  //processed by the Gateway daemon and CICS.
                  if (rc==RC_OK &&
                     runTxn.getRc()==ECIRequest.ECI_NO_ERROR) {
                     requestTimings.success();
                     displayOutput.successfulRequest(requestTimings);
                  } else {  
                     requestTimings.fail();
                     displayOutput.failedRequest();
                     if (runTxn.getRc()!=ECIRequest.ECI_NO_ERROR) {
                        if (runTxn.getRc()==ECIRequest.ECI_ERR_TRANSACTION_ABEND) {
                           ErrorStore.getInstance().set(runTxn.getRcString()+" ("+
                                  runTxn.Abend_Code+")");
                        } else {
                           ErrorStore.getInstance().set(runTxn.getRcString());
                        }
                     }
                     rc = RC_REQUEST_FAIL;               
                  }
                  if (isNotLastRequestInLoop(requestLoop)) {
                     displayOutput.delimiterBetweenRequests();
                  }
               }
               displayOutput.endOfRequest();

            }

            if (opened) {
               boolean closeSuccess;
               closeTimings.setStart();
               try {
                  jg.close();  
                  closeSuccess=true;
               } catch (Exception e) {
                  closeSuccess=false;
                  closeTimings.fail();                
                  if (rc == RC_OK) {
                     // We always call close even if a previous call failed to make
                     // sure the connection is no longer open. Only set an error
                     // if there wasn't one already, to prevent us overwriting the
                     // first error condition.
                     rc = CLOSE_FAIL;
                  }
               }
               closeTimings.setEnd();

               if (closeSuccess) {
                  closeTimings.success();
                  displayOutput.successfulClose(closeTimings);
               }
                           
            }

            try {
               Thread.sleep(thinkTime.getValueAsInt());
            } catch (InterruptedException e) {
               e.printStackTrace();
            }

         }
      }

      // If we had any request errors, output the text of the first error
      if (ErrorStore.getInstance().hasErrorOccurred()) {
         displayOutput.error();
      }

      displayOutput.summaryTitle();

      displayOutput.outputTotals("Opens", openTimings);
      displayOutput.outputTotals("Requests", requestTimings);
      displayOutput.outputTotals("Closes", closeTimings);

      return rc;
   }

   private boolean isNotLastRequestInLoop(int requestLoop) {
      return requestLoop + 1 < numberRequests.getValueAsInt();
   }

   private JavaGateway setupJavaGateway() throws IOException {
      JavaGateway jg = new JavaGateway();

      jg.setURL(gatewayURL.getURL());
      if (gatewayURL.isPortSpecified()) {
         try {
            jg.setPort(gatewayURL.getPort());
         } catch (PortNotSpecifiedException e) {
            // We should never get to this point.
            // Store the error string and fail the method if we do.
            ErrorStore.getInstance().set(e);
            throw new IOException(e.getMessage());
         }
      }
      return jg;
   }

   private int parseURL(String targetURL) {
      String url = null;
      String port = null;
      int rc = RC_OK;

      int col = targetURL.lastIndexOf(':');
      if (targetURL.lastIndexOf(':', col - 1) == -1) {
         // The URL is of the form 'protocol://url:port
         // If we only have one ':' instance it means that there is no port
         // specified.
         url = targetURL;
         port = null;
      } else {
         url = targetURL.substring(0, col);
         port = targetURL.substring(col + 1);
      }

      int portint = 0;
      if (port != null) {
         try {
            portint = Integer.parseInt(port);
         } catch (NumberFormatException e) {
            System.out.println("A numeric value is required for the port");
            displayOutput.displayUsage();
            rc = INVALID_PARAMETERS;
         }
      }

      if (port != null) {
         gatewayURL = new URLSpecifier(url, portint);
      } else {
         gatewayURL = new URLSpecifier(url);
      }

      return rc;
   }

   private byte[] createPayload(int payloadSize) {
      byte payload[] = new byte[payloadSize];
      // Putting a terminating 'X' prevents the commarea being null stripped
      // without the CPU overhead of requiring ctgping to overwrite the whole
      // data area
      payload[payloadSize - 1] = 'X';

      return payload;
   }

   public ErrorStore getMyError() {
      return ErrorStore.getInstance();
   }

   public DisplayOut getDisplayOutput() {
      return displayOutput;
   }

}
